#include <map>
#include <vector>
#include <string>
#include <iostream>
#include <unordered_map>

#include "leveldb/db.h"

using namespace std;

#define FP_SIZE 6

/* chunks and their neighbor_chunks */
struct node 
{
    uint64_t count; // chunk frequency
    unordered_map<string, uint64_t> leftmap; // left_neighbors
    unordered_map<string, uint64_t> rightmap;  // right_neighbors

};
unordered_map<string, node> mymap;

typedef pair<string, node> hash_pair;
typedef pair<string, uint64_t> neighbor_pair;

leveldb::DB *frequency_db; // db of frequency of chunks
leveldb::DB *left_db; // db of left_neighbor_chunks's co-occurrence frequency of chunks
leveldb::DB *right_db; // db of right_neighbor_chunks's co-occurrence frequency of chunks


/* init db */
void init_db(char *db_name, leveldb::DB **db);

/* read hash_file to count frequency of chunks and co-occurrence frequency of neighbor_chunks */
void count_hashes_to_DB(FILE *fp);

/* count frequency of chunks: key->chunkhash, value->reference_count */
void count_node_to_Frequency_DB(char * tmp_hash);

/* count co-occurrence frequency of neighbor_chunks: key->chunkhash, value->neighbor_hash+neighbor_count.... */
void count_neighbor_node_to_DB(int flag, char * tmp_current_hash, char * tmp_neighbor_hash);

/* store in leveldb */
void storage();

/* print chunk hash */
void print_fp(char * a);

void print_fp(char * a)
{
    int len = 0;
    for (len = 0; len < FP_SIZE; len++) {
        printf("%02x", (unsigned char)a[len]);
        if (len < FP_SIZE - 1) printf(":");
    }
    printf("\n");
}

int main(int argc, char *argv[])
{
    // argv[1] points to hash file;
    // argv[2] points to frequency_db;
    // argv[3] points to left_db
    // argv[4] points to right_db;
    if (argc != 5) {
        printf("[Error] the number of parameters should be 5!\n");
        return -1;
    }

    init_db(argv[2], &frequency_db);
    init_db(argv[3], &left_db);
    init_db(argv[4], &right_db);

    FILE *fp = fopen(argv[1], "r");
    if (fp == nullptr) {
        printf("[Error] Failed to open %s\n", argv[1]);
        return -1;
    }

    count_hashes_to_DB(fp);
    storage();

    fclose(fp);
    delete frequency_db;
    delete left_db;
    delete right_db;
    return 0;
}

/* init db */
void init_db(char *db_name, leveldb::DB **db)
{
    leveldb::Options options;
    options.create_if_missing = true;
    leveldb::Status status = leveldb::DB::Open(options, db_name, db);
    if (!status.ok()) {
        printf("[Error] Failed to open DB!\n");
        exit(-1);
    }
    if (db == nullptr) {
        printf("[Error] Failed to assign DB pointer!\n");
        exit(-1);
    }
}

/* read hash_file to count frequency of chunks and co-occurrence frequency of neighbor_chunks */
void count_hashes_to_DB(FILE *fp)
{
    char read_buffer[256];
    char *item;
    bool first_node_flag = true;

    // store previous chunk hash value
    char previous_hash[FP_SIZE+1];
    memset(previous_hash, 0, FP_SIZE+1);

    // a new chunk
    char current_hash[FP_SIZE+1];
    // Read hash value from hash_file generated by `fs-hasher`
    while (fgets(read_buffer, 256, fp)) {
        // skip title line and find next file
        if (strpbrk(read_buffer, "Chunk")) continue;
        if (strpbrk(read_buffer, "NEXTFILE"))  {
            first_node_flag =  true;
            continue;
        }
        // reset to a brand-new chunk hash value
        memset(current_hash, 0, FP_SIZE+1);

        // 1. Read hash of chunk
        // the separator below works correctly
        item = strtok(read_buffer, ":\t\n ");
        int idx = 0;
        while (item != NULL && idx < FP_SIZE) {
            // store hex value of hash into char array with base 16
            current_hash[idx] = strtol(item, NULL, 16);
            idx++;
            item = strtok(NULL, ":\t\n");
        }
        current_hash[FP_SIZE]='\0';

        //---------------------------start counting into frequency db-----------------------------
        // 2. Count to frequency_db
        // Node mapping structure: (Key, Value) => (hash, count)
        count_node_to_Frequency_DB(current_hash);
        //---------------------------Finish counting into frequency db-----------------------------

        if (!first_node_flag) {
            //---------------------------start counting into left db-----------------------------
            // Left: 1. Set left hash of the current hash
            // Example: C1, C2, C5, C2
            //          set C1 hash as previous_hash_value since C1 is the left node of C2

            // 3. Count to left_db
            count_neighbor_node_to_DB(0 , current_hash, previous_hash);
            //---------------------------Finish counting into left db-----------------------------

            //---------------------------start counting into right db-----------------------------
            // Count right nodes into Right_DB, using the same method as in counting left nodes to Left_DB

            // Right: Consider the left node as the current node and the current node as the right node of the left one
            // Example: C1, C2(Current Node), C5, C2
            //          - Consider C1 as the current node, aka, `fake current node` in code below
            //          - Consider C2 as the right node of the C1

            // 4. Count to right_db
            // right_node is the right node of `fake current node` also set the actual value of the current node to it
            // leveldb::Slice fake_current(previous_hash, FP_SIZE);
            count_neighbor_node_to_DB(1 , previous_hash, current_hash);
            //---------------------------Finish counting into right db-----------------------------
        }

        // 4. Update previous_hash chunk
        memcpy(previous_hash, current_hash, FP_SIZE+1);
        if (first_node_flag) first_node_flag = false;
    }

    //-------------------------------------counting finished---------------------------------------
}

/* count frequency of chunks: key->chunkhash, value->reference_count */
void count_node_to_Frequency_DB(char * tmp_hash) {
    string current_hash;
    current_hash.resize(FP_SIZE);
    current_hash.assign(tmp_hash,FP_SIZE);
    
    uint64_t count;
    unordered_map<string, node> ::const_iterator it;
    
    unordered_map<string, node> tmp;
    struct node tmpnode;
    
    if(mymap.size()==0){
        count = 1; 
        mymap[current_hash].count = count;
        return;
    }

    it = mymap.find(current_hash);
    if(it != mymap.end()) {
        count = it->second.count;
        count++;
        mymap[current_hash].count = count;

    } else {
        count = 1; 
        tmpnode.count = count;
        mymap[current_hash] = tmpnode;
    }

}

/* count co-occurrence frequency of neighbor_chunks: key->chunkhash, value->neighbor_hash+neighbor_count.... */
void count_neighbor_node_to_DB(int flag, char * tmp_current_hash, char * tmp_neighbor_hash) {
    string current_hash;
    current_hash.resize(FP_SIZE);
    current_hash.assign(tmp_current_hash,FP_SIZE);

    string neighbor_hash;
    neighbor_hash.resize(FP_SIZE);
    neighbor_hash.assign(tmp_neighbor_hash,FP_SIZE);

    uint64_t count;
    unordered_map<string, node> ::const_iterator it;
    
    if(flag == 0) { // left_neighbor 
        it = mymap.find(current_hash);
        if(it == mymap.end())printf("error\n");
        
        unordered_map<string, uint64_t> ::const_iterator nit;
        
        if(it->second.leftmap.size() == 0) {
            count = 1;
            mymap[current_hash].leftmap[neighbor_hash]= count;
            return;
        }

        nit = it->second.leftmap.find(neighbor_hash);
        if(nit != it->second.leftmap.end()) {
            count = nit->second;
            count++;
            mymap[current_hash].leftmap.erase(neighbor_hash);
        } else {
            count = 1; 
        }
        mymap[current_hash].leftmap[neighbor_hash] = count;

    } 
    else if(flag == 1) { // right_neighbor 
        it = mymap.find(current_hash);
        if(it == mymap.end())printf("error\n");
        
        unordered_map<string, uint64_t> ::const_iterator nit;
        
        if(it->second.rightmap.size() == 0) {
            count = 1;
            mymap[current_hash].rightmap[neighbor_hash]= count;
            return;
        }

        nit = it->second.rightmap.find(neighbor_hash);

        if(nit != it->second.rightmap.end()) {
            count = nit->second;
            count++;
            mymap[current_hash].rightmap.erase(neighbor_hash);
        } else {
            count = 1; 
        }
        mymap[current_hash].rightmap[neighbor_hash] = count;
    }

}

/* store in leveldb */
void storage() {
    
    uint64_t count;
    unordered_map<string, node> ::const_iterator it;
    
    unordered_map<string, uint64_t> tmp;
    struct node tmpnode;
    
    for(it = mymap.begin(); it != mymap.end(); it++) { 
        
        leveldb::Status status;
        leveldb::Slice key_slice(it->first.c_str(), FP_SIZE);

        string count_buf;
        
        uint64_t tmpcount = it->second.count;
        count_buf.clear();
        count_buf.resize(sizeof(uint64_t));
        count_buf.assign((char *)&tmpcount, sizeof(uint64_t));
        

        leveldb::Slice update_slice1(count_buf.c_str(), sizeof(uint64_t));
        status = frequency_db->Put(leveldb::WriteOptions(), key_slice, update_slice1);
        if (status.ok() == 0) fprintf(stderr, "[Error] %s\n", status.ToString().c_str());

        unordered_map<string, uint64_t> ::const_iterator nit;
        
        // left_db
        string neighbor_buf_left;
        neighbor_buf_left.clear();

        for(nit = it->second.leftmap.begin(); nit != it->second.leftmap.end(); nit++) {
            count_buf.clear();
            count_buf.resize(FP_SIZE);
            count_buf.assign((char *)nit->first.c_str(), FP_SIZE);
            neighbor_buf_left += count_buf;

            tmpcount=nit->second;
            count_buf.clear();
            count_buf.resize(sizeof(uint64_t));
            count_buf.assign((char *)&tmpcount, sizeof(uint64_t));
            neighbor_buf_left += count_buf;
        }
        neighbor_buf_left += '\0';
        leveldb::Slice update_slice2(neighbor_buf_left.c_str(), neighbor_buf_left.size());
        status = left_db->Put(leveldb::WriteOptions(), key_slice, update_slice2);
        if (status.ok() == 0) fprintf(stderr, "[Error] %s\n", status.ToString().c_str());

        // right_db
        string neighbor_buf_right;
        neighbor_buf_right.clear();

        for(nit = it->second.rightmap.begin(); nit != it->second.rightmap.end(); nit++) {
            count_buf.clear();
            count_buf.resize(FP_SIZE);
            count_buf.assign((char *)nit->first.c_str(), FP_SIZE);
            neighbor_buf_right += count_buf;

            tmpcount = nit->second;
            count_buf.clear();
            count_buf.resize(sizeof(uint64_t));
            count_buf.assign((char *)&tmpcount, sizeof(uint64_t));
            neighbor_buf_right += count_buf;
        }
        neighbor_buf_right += '\0';
        leveldb::Slice update_slice3(neighbor_buf_right.c_str(), neighbor_buf_right.size());
        status = right_db->Put(leveldb::WriteOptions(), key_slice , update_slice3);
        if (status.ok() == 0) fprintf(stderr, "[Error] %s\n", status.ToString().c_str());
    }
}
